from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import

from .errors import ElementNotFoundError, ElementError
from .context.expression import Expression
from . import namespaces
from .compat import (text_type,
                     implements_to_string,
                     iteritems,
                     pickle)

from fs.path import dirname, pathjoin

import re
import weakref
from collections import defaultdict

_re_xml_namespace = re.compile(r'^(?:\{(.*?)\})*(.*)$', re.UNICODE)


def _extract_namespace(tag_name):
    """Extracts namespace and tag name in Clark's notation"""
    return _re_xml_namespace.match(tag_name).groups()


def _childless_tag(tag_name, attr_map):
    attrs = _get_attrs(attr_map)
    if attrs:
        a = ' '.join('%s="%s"' % (k, v) for k, v in sorted(attrs.items())).strip()
        return '<%s>' % ' '.join((tag_name, a))
    return '<%s>' % tag_name


def _get_attrs(attr_map):
    attrs = {}
    for ns, attr_map in attr_map.items():
        if ns == namespaces.default:
            ns = ''
        for k, v in iteritems(attr_map):
            if ns:
                attr_name = "{{{}}}{}".format(ns, k)
            else:
                attr_name = k
            attrs[attr_name] = v
    return attrs


@implements_to_string
class DocumentStructure(object):
    def __init__(self, document, library, xml):
        self.document = document
        self.library = library
        self.xml = xml
        self.doc_id = 1
        self.nodes = {}
        self.docid_counts = defaultdict(int)
        if library is not None:
            self.lib_long_name = library.long_name
        else:
            self.lib_long_name = None

    def __str__(self):
        return text_type(self.document)

    @property
    def top_nodes(self):
        root = self.nodes[1]
        return root.children

    @property
    def root_node(self):
        return self.nodes[1]

    def dumps(self):
        serialized = {
            "xml": self.xml,
            "nodes": self.nodes
        }
        return pickle.dumps(serialized, pickle.HIGHEST_PROTOCOL)

    @classmethod
    def load(cls, doc_dump, document, library):
        serialized = pickle.loads(doc_dump)
        structure = DocumentStructure(document, library, serialized["xml"])
        nodes = serialized['nodes']
        for node in nodes.itervalues():
            node.structure = structure
        structure.nodes = nodes
        return structure

    @classmethod
    def quick_load(cls, document, library, xml, nodes):
        structure = DocumentStructure(document, library, xml)
        for node in nodes.itervalues():
            node.structure = structure
        structure.nodes = nodes
        return structure

    def add_node(self, node):
        doc_id = self.doc_id
        self.doc_id += 1
        self.nodes[doc_id] = node
        node.doc_id = doc_id
        if node.parent_doc_id is not None:
            self.nodes[node.parent_doc_id].children.append(node)
        node.structure = self
        node.lib_long_name = self.lib_long_name

    def render(self):
        def write_node(node, level=0):
            tab = "    " * level
            if isinstance(node, DocumentTextNode):
                print("{}{!r}".format(tab, node.text))
            else:
                print("{}{}".format(tab, node))
                for child in node.children:
                    write_node(child, level + 1)
        write_node(self.nodes[1])


@implements_to_string
class DocumentNode(object):
    def __init__(self,
                 xmlns,
                 tag_name,
                 parent_doc_id,
                 attrs,
                 translatable_attrs,
                 text,
                 source_line,
                 translate_text=False):
        self.text_node = False
        self.xmlns = xmlns
        self.tag_name = tag_name
        self.tag_type = (xmlns, tag_name)
        self.parent_doc_id = parent_doc_id
        self.attrs = attrs
        self.translatable_attrs = translatable_attrs
        self.source_line = source_line
        self.text = text
        self.translate_text = translate_text

        self.children = []
        self.structure = None
        self.text_nodes = None
        self.element = None
        self.lib_long_name = None

    def __repr__(self):
        return _childless_tag(self.xmlns_name, self.attrs)

    def __str__(self):
        return _childless_tag(self.xmlns_name, self.attrs)

    def __getstate__(self):
        state = self.__dict__.copy()
        state.pop('structure')
        return state

    @property
    def parent(self):
        if self.parent_doc_id:
            return self.structure.nodes[self.parent_doc_id]

    def build(self, archive, context):
        """Build an element from structure data"""
        # Returns element, or False if element could not be yet, or None if no element is required
        attrs = self.attrs.get(namespaces.default, {})
        docname = attrs.get('docname', None)
        libname = attrs.get('libname', None)
        structure = self.structure
        document = structure.document
        tag_name = self.tag_name

        structure.docid_counts[tag_name] += 1
        docid = "{}.{}".format(tag_name, structure.docid_counts[tag_name])

        if libname is None and structure.library is not None:
            libname = structure.library.allocate_libname(docid)

        replace_lib = False
        if libname is not None and '#' in libname:
            lib_, _, libname = libname.partition('#')
            lib = archive.get_or_create_library(lib_)
            replace_lib = True
        else:
            lib = self.structure.library

        element_type = archive.registry.get_element_type(self.xmlns, self.tag_name)
        if element_type is None:
            # We don't know how to build this yet
            # May be defined in a document not yet built
            return False

        if self.parent and self.parent.element:
            parent_docid = self.parent.element.docid
        else:
            parent_docid = None
        element = element_type(structure.document,
                               self.xmlns,
                               tag_name,
                               parent_docid,
                               docid,
                               source_line=self.source_line)
        element.libname = libname
        element.docname = docname
        element._code = structure.xml
        element._build(context, self.text or '', attrs, self.translatable_attrs)
        element._document = weakref.ref(document)

        element._let_exp = None  # Cache for let expressions
        element._let = self.attrs.get(namespaces.let, None)
        self.element = element

        element._translate_text = element._meta.translate or self.translate_text

        if element.parent:
            self.text_nodes = element._meta.text_nodes or element.parent._meta.text_nodes

        if docname is not None:
            document.register_named_element(docname, element)

        if lib:
            element._libid = "{}#{}".format(lib.long_name, libname)
            if replace_lib:
                lib.add_replacement_node(element)
            else:
                lib.register_named_element(libname,
                                           element,
                                           priority=structure.library.priority)

        return element

    @property
    def xmlns_name(self):
        if self.xmlns == namespaces.default:
            return self.tag_name
        else:
            return "{{{}}}{}".format(self.xmlns, self.tag_name)

    @property
    def tag_display_name(self):
        return "<{}>".format(self.xmlns_name)


class DocumentTextNode(DocumentNode):
    def __init__(self,
                 parent_doc_id,
                 source_line,
                 text):
        self.text_node = True
        self.parent_doc_id = parent_doc_id
        self.source_line = source_line
        self.text = text
        self.attrs = {}
        self.translatable_attrs = ()
        self.children = []
        self.translate_text = False

    def __repr__(self):
        return repr(self.text)

    def build(self, archive, context):
        text_nodes = self.parent.text_nodes
        if not text_nodes:
            return None
        xmlns, tag_name = _extract_namespace(self.parent.text_nodes)
        self.xmlns = xmlns or namespaces.default
        self.tag_name = tag_name
        self.tag_type = (self.xmlns, tag_name)
        element = super(DocumentTextNode, self).build(archive, context)
        if self.parent.translate_text:
            element.translate = True
        return element


class Document(object):

    def __init__(self, archive, lib=None, path=None):
        self._archive = weakref.ref(archive)
        self.elements = {}
        self.named_elements = {}
        self.root_element = None
        self.lib = lib
        self.fs = None
        self.path = path
        self.structure = None

        self.finalized = False
        self.document_finalized = False
        self.lib_finalized = False

    @property
    def archive(self):
        return self._archive()

    def __repr__(self):
        if self.lib:
            return "<document %s:\"%s\">" % (self.lib.long_name, self.path)
        else:
            return '<document "%s">' % self.path

    def __getitem__(self, element_id):
        return self.elements[element_id]

    def dumps(self, expressions=None):
        if expressions is None:
            expressions = []
        doc = {"expressions": expressions,
               "root": self.root_element.dumps()}
        return doc

    def loads(self, data):
        from .elements.elementbase import ElementBase
        Expression.insert_expressions(data["expressions"])
        self.root_element = ElementBase.loads(data["root"], self)
        return self

    def show(self):
        print()
        print(self)

        def recurse(node, level=0):
            print("{}{}".format(" " * (level * 4), node))
            for child in node._children:
                recurse(child, level + 1)
        if self.root_element:
            recurse(self.root_element)

    def get_element(self, element_ref, app=None, lib=None):
        try:
            if '#' not in element_ref:
                return app, self.named_elements[element_ref]
        except KeyError:
            raise ElementNotFoundError(element_ref, app=app, lib=lib)
        return self.archive.get_element(element_ref, app=app, lib=lib)

    def get_app_element(self, element_ref, app=None):
        if '#' in element_ref:
            app_id, lib_id, name = self.archive.parse_element_ref(element_ref)
            if lib_id is not None:
                try:
                    app = self.archive.find_app(lib_id)
                except Exception as e:
                    raise ElementNotFoundError(element_ref, app=app, reason=text_type(e))
                element_ref = "{}#{}".format(app.name, name)
        return self.get_element(element_ref, app=app)

    def detect_app_element(self, context, element_ref, app=None):
        if '#' in element_ref:
            app_id, lib_id, name = self.archive.parse_element_ref(element_ref)
            if lib_id is not None:
                try:
                    app = self.archive.detect_app(context, lib_id)
                except Exception as e:
                    raise ElementNotFoundError(element_ref, app=app, reason=text_type(e))
                element_ref = "{}#{}".format(app.name, name)
        return self.get_element(element_ref, app=app)

    def qualify_element_ref(self, element_ref, app=None, lib=None):
        if element_ref.find('#', 1) != -1:
            return element_ref
        app, element = self.get_element(element_ref, app=app, lib=lib)
        if element is None:
            raise ElementNotFoundError(element_ref, app=app, lib=lib)
        if self.lib:
            if app is not None:
                qname = app.name
            else:
                qname = self.lib.long_name
            return "%s#%s" % (qname, element.libname)
        return element_ref

    qualify = qualify_element_ref

    def get_root(self):
        return self.root_element
    root = property(get_root)

    def register_element(self, element):
        self.elements[element.docid] = element
        if element.parent_docid is not None:
            self.elements[element.parent_docid]._add_child(element)
        else:
            self.root_element = element

    def register_named_element(self, docname, element):
        if docname in self.named_elements:
            raise ElementError("docname '{}' is already present in this document".format(docname),
                               element=element)
        self.named_elements[docname] = element

    def get_named_element(self, docname):
        return self.named_elements.get(docname, None)

    def iter_elements(self):
        return self.elements.iter_values()

    def __iter__(self):
        return self.iter_elements()

    def resolve_relative_path(self, path):
        document_dirname = dirname(self.path)
        new_path = pathjoin(document_dirname, path)
        return new_path

    def finalize(self, context):
        if self.finalized:
            return
        ignore_errors = bool(context['._ignore_finalize_errors'])

        def do_finalize(element):
            for child in element:
                do_finalize(child)
            try:
                element.finalize(context)
            except:
                if not ignore_errors:
                    raise
        if self.root:
            do_finalize(self.root)
        self.finalize = True

    def document_finalize(self, context):
        if self.document_finalized:
            return
        ignore_errors = bool(context['._ignore_finalize_errors'])

        def do_finalize(element):
            for child in element:
                do_finalize(child)
            try:
                element.document_finalize(context)
            except:
                if not ignore_errors:
                    raise
        if self.root:
            do_finalize(self.root)
        self.document_finalized = True

    def lib_finalize(self, context):
        if self.lib_finalized:
            return
        ignore_errors = bool(context['._ignore_finalize_errors'])

        def do_finalize(element):
            for child in element._children:
                do_finalize(child)
            try:
                element.lib_finalize(context)
            except:
                if not ignore_errors:
                    raise
        if self.root_element:
            do_finalize(self.root_element)
        self.lib_finalized = True



from __future__ import unicode_literals
from __future__ import print_function

from .errors import ElementNotFoundError

from .compat import text_type
import logging
log = logging.getLogger('moya.db')


class GenericKey(object):
    """A generic foreign key"""
    def __init__(self, app, model, pk):
        self.app = app
        self.model = model
        self.pk = pk

    def __repr__(self):
        if self.app is None:
            return "<generic-key (null)>"
        return "<generic-key {}#{} {}>".format(self.app.name, self.model.libname, self.pk)

    @classmethod
    def from_object(cls, obj):
        if obj is None:
            return cls(None, None, None)
        return cls(obj._app, obj._model, obj.id)

    @classmethod
    def decode(cls, value):
        if not value:
            return cls(None, None, None)
        try:
            app, model, pk = [v or None for v in value.split(',')]
            pk = int(pk)
        except ValueError:
            # Malformed generic key in the database
            return cls(None, None, None)
        from moya import pilot
        # TODO: is there a better way of getting the archive here?
        archive = pilot.context['.app'].archive
        app = archive.get_app(app)
        if app is None:
            # No such app, treat it as a null
            return cls(None, None, None)

        try:
            app, model = archive.get_element('{}#{}'.format(app.name, model))
        except:
            # Model doesn't exist, treat it as a null
            return cls(None, None, None)

        any_key = cls(app, model, pk)
        return any_key

    def __moyacontext__(self, context):
        return self.lookup(context)

    def __moyadbobject__(self):
        return self.encode()

    def encode(self):
        if not self.app:
            return None
        return ",".join((self.app.name, self.model.libname, text_type(self.pk)))

    def lookup(self, context):
        if self.app is None:
            return None

        element_ref = '{}#{}'.format(self.app.name, self.model.libname)
        try:
            model_app, model = self.app.archive.get_element(element_ref)
        except ElementNotFoundError:
            log.warning('no model found for generic key {!r}', self)
            return None
        db = model.get_db()
        dbsession = context['._dbsessions'][db.name]
        table_class = model.get_table_class(model_app)
        qs = dbsession.query(table_class).filter(table_class.id == self.pk)
        return qs.first()



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import

from .build import build_server
from . import errors
from . import pilot
from . import db
from .context import Context
from . import tags
from .tags import cookie
from .tools import timer, lazystr
from .logic import debug_lock, is_debugging
from .logic import notify
from .request import MoyaRequest, ReplaceRequest
from .response import MoyaResponse
from . import http
from .compat import text_type, itervalues, py2bytes
from . import namespaces
from .loggingconf import init_logging_fs
from .context.expression import Expression


from webob import Response

from fs.path import splitext
from fs.opener import fsopendir
from fs.errors import FSError

import random
from time import time, clock, sleep
from threading import RLock
import weakref
from collections import defaultdict
from textwrap import dedent
import os.path

import logging
log = logging.getLogger("moya")
request_log = logging.getLogger("moya.request")
startup_log = logging.getLogger("moya.startup")
preflight_log = logging.getLogger("moya.preflight")

try:
    import watchdog
    import watchdog.events
    import watchdog.observers
except ImportError:
    watchdog = None

if watchdog:
    class ReloadChangeWatcher(watchdog.events.FileSystemEventHandler):

        def __init__(self, watch_fs, app):
            self._app = weakref.ref(app)
            self.watch_types = app.archive.cfg.get_list("autoreload", "extensions", ".xml\n.ini\n.py")
            self.watching_fs = watch_fs
            self.observer = None
            try:
                path = self.watching_fs.getsyspath('/')
            except FSError:
                startup_log.warning('auto reload not available on this filesystem')
            else:
                try:
                    observer = watchdog.observers.Observer()
                    observer.schedule(self, path, recursive=True)
                    observer.start()
                    self.observer = observer
                except:
                    startup_log.exception('failed to watch "{}" for changes'.format(path))
                else:
                    startup_log.debug('watching "{}" for changes'.format(path))

            super(ReloadChangeWatcher, self).__init__()

        def on_any_event(self, event):
            ext = splitext(event.src_path)[1].lower()
            if ext not in self.watch_types:
                return

            if not self.app.rebuild_required:
                log.info("detected modification to project, rebuild will occur on next request")
                self.app.rebuild_required = True

        @property
        def app(self):
            return self._app()

        def close(self):
            if self.observer is not None:
                try:
                    self.observer.stop()
                    # No need to join since we are exiting the process
                    # self.observer.join()
                except:
                    pass
            self.watching_fs.close()

else:
    class ReloadChangeWatcher(object):
        def __init__(self, watch_fs, app):
            startup_log.warning("'watchdog' module could not be imported, autoreload is disabled")
            startup_log.warning("you might be able to fix this with 'pip install watchdog'")

        def close(self):
            pass


class WSGIApplication(object):

    def __init__(self,
                 filesystem_url,
                 settings_path,
                 server="main",
                 logging=None,
                 disable_autoreload=False,
                 breakpoint=False,
                 breakpoint_startup=False,
                 validate_db=False,
                 simulate_slow_network=False,
                 strict=False,
                 master_settings=None,
                 test_build=False,
                 develop=False,
                 load_expression_cache=True):
        self.filesystem_url = filesystem_url
        self.settings_path = settings_path
        self.server_ref = server
        self.logging = logging
        self.breakpoint = breakpoint
        self.validate_db = validate_db
        self.watching_fs = None
        self.rebuild_required = False
        self._new_build_lock = RLock()
        self.archive = None
        self._self = weakref.ref(self, self.on_close)
        self.simulate_slow_network = simulate_slow_network
        self.master_settings = master_settings
        self.test_build = test_build
        self.develop = develop
        self.load_expression_cache = load_expression_cache

        if logging is not None:
            with fsopendir(self.filesystem_url) as logging_fs:
                init_logging_fs(logging_fs, logging)
        try:
            self.build(breakpoint=breakpoint_startup, strict=strict)
        except Exception as e:
            startup_log.critical(text_type(e))
            raise
        self.watcher = None
        if self.archive.auto_reload and not disable_autoreload:
            try:
                location = self.archive.project_fs.getsyspath('/')
            except FSError:
                log.warning('project filesystem has no syspath, disabling autoreload')
            else:
                watch_location = os.path.join(location, self.archive.cfg.get('autoreload', 'location', ''))
                self.watcher = ReloadChangeWatcher(fsopendir(watch_location), self)

    @classmethod
    def on_close(cls, application_weakref):
        # Called prior to Python finalizing the WSGIApplication, but before __del__
        # Note, application_weakref will always return None. There is no way to use the original object at this point
        pass

    def close(self):
        if self.watcher is not None:
            self.watcher.close()

    def __repr__(self):
        return """<wsgiapplication {} {}>""".format(self.settings_path, self.server_ref)

    def build(self, breakpoint=False, strict=False):
        with timer('startup', output=startup_log.debug):
            build_result = build_server(self.filesystem_url,
                                        self.settings_path,
                                        server_element=self.server_ref,
                                        validate_db=self.validate_db,
                                        breakpoint=breakpoint,
                                        strict=strict,
                                        master_settings=self.master_settings,
                                        test_build=self.test_build,
                                        develop=self.develop)
        if build_result is None:
            msg = "Failed to build project"
            raise errors.StartupFailedError(msg)

        self.archive = build_result.archive
        self.archive.finalize()
        self.server = build_result.server

        if self.load_expression_cache:
            if self.archive.has_cache('parser'):
                parser_cache = self.archive.get_cache('parser')
                if Expression.load(parser_cache):
                    log.debug('expression cache loaded')

        context = Context({"console": self.archive.console,
                           "settings": self.archive.settings,
                           "debug": self.archive.debug,
                           "develop": self.develop or self.archive.develop,
                           "pilot": pilot})

        self.populate_context(context)
        self.archive.populate_context(context)
        self.archive.fire(context, "sys.startup")
        db.commit_sessions(context)

    def populate_context(self, context):
        # Called by moya <command>
        context.root.update(_dbsessions=db.get_session_map(self.archive),
                            console=self.archive.console,
                            fs=self.archive.get_context_filesystems())

    def do_rebuild(self):
        self.archive.console.div("Re-building project due to changes", bold=True, fg="blue")

        error_text = None
        try:
            new_build = build_server(self.filesystem_url,
                                     self.settings_path,
                                     server_element=self.server_ref,
                                     strict=self.archive.strict,
                                     validate_db=True)
        except Exception as e:
            error_text = text_type(e)
            log.warning(e)
            new_build = None

        if new_build is None:
            self.rebuild_required = False
            notify("Rebuild Failed",
                   error_text or "Unable to build project, see console")
            return

        with self._new_build_lock:
            self.archive = new_build.archive
            self.server = new_build.server
            self.archive.finalize()
        import gc
        gc.collect()
        self.rebuild_required = False
        self.archive.console.div("Modified project built successfully", bold=True, fg="green")

    def preflight(self, report=True):
        app_preflight = []

        if self.archive.preflight:
            for app in itervalues(self.archive.apps):
                preflight = []
                for element in app.lib.get_elements_by_type((namespaces.preflight, "check")):
                    preflight_callable = self.archive.get_callable_from_element(element, app=app)

                    context = Context({"preflight": preflight})
                    self.archive.populate_context(context)
                    self.populate_context(context)
                    context['.app'] = app

                    if not element.check(context):
                        preflight.append((element, "skip", ''))
                        continue

                    try:
                        preflight_callable(context, app=app)
                    except Exception as e:
                        preflight.append((element, "error", text_type(e)))

                app_preflight.append((app, preflight))

            if report:
                all_ok = True
                for app, checks in app_preflight:
                    if not checks:
                        continue
                    totals = defaultdict(int)
                    for element, status, text in checks:
                        lines = dedent(text).splitlines()
                        totals[status] += 1
                        for line in lines:
                            if line:
                                if status == "warning":
                                    preflight_log.warning('%s', line)
                                elif status == "fail":
                                    preflight_log.error('%s', line)
                                elif status == "error":
                                    preflight_log.critical('%s', line)

                    results = []
                    for status in ("warning", "fail", "error"):
                        if totals[status]:
                            results.append("{} {}(s)".format(totals[status], status))
                            if status != 'skip':
                                all_ok = False
                    if results:
                        preflight_log.info("%s in %s", ", ".join(results), app)

                if all_ok:
                    preflight_log.info("all passed")
                else:
                    preflight_log.warning("preflight detected potential problems -- run 'moya preflight' for more information")

        return app_preflight

    def get_response(self, request, context):
        """Get a response object"""
        fire = self.archive.fire
        fire(context, "request.start", app=None, sender=None, data={'request': request})

        with pilot.manage_request(request, context):

            root = context.root
            root.update(settings=self.archive.settings,
                        debug=self.archive.debug,
                        request=request,
                        cookiejar=cookie.CookieJar())
            self.populate_context(context)

            fire(context, "request.pre-dispatch", data={'request': request})
            while 1:
                try:
                    result = self.server.dispatch(self.archive, context, request, breakpoint=self.breakpoint)
                except Exception:
                    log.exception("error in dispatch")
                    raise

                if isinstance(result, ReplaceRequest):
                    context.root['request'] = request = result.request
                    continue
                break

            fire(context, "request.post-dispatch", data={'request': request, 'result': result})

            response = None
            if result is not None:
                if isinstance(result, text_type):
                    response = MoyaResponse(charset=py2bytes('utf8'),
                                            text=text_type(result))
                elif isinstance(result, Response):
                    response = result
            else:
                response = context.root.get("response", None)

            if response is None:
                response = MoyaResponse(status=http.StatusCode.not_found,
                                        text=py2bytes("404 - Not Found"))

            if 'headers' in root:
                for k, v in root['headers'].items():
                    response.headers[k.encode('utf-8')] = v.encode('utf-8')

        fire(context, "request.response", data={'request': request, 'response': response})

        return response

    def __call__(self,
                 environ,
                 start_response):
        """Build the request"""

        if self.rebuild_required and not is_debugging():
            with debug_lock:
                self.do_rebuild()

        if self.simulate_slow_network:
            sleep(random.uniform(.2, .5))

        start = time()
        start_clock = clock()
        context = Context()
        request = MoyaRequest(environ)

        response = self.get_response(request, context)
        taken = time() - start
        clock_taken = clock() - start_clock

        start_response(response.status, response.headerlist)

        taken_ms = lazystr("{:.1f}ms {:.1f}ms".format, taken * 1000, clock_taken * 1000)
        request_log.info('"%s %s %s" %i %s %s',
                         request.method,
                         request.path_qs,
                         request.http_version,
                         response.status_int,
                         response.content_length or 0,
                         taken_ms)

        if request.method == 'HEAD':
            return []
        else:
            return response.app_iter

        self.archive.fire(context,
                          "request.end",
                          data={'response': response})

Application = WSGIApplication



#!/usr/bin/env python

from setuptools import setup, find_packages

VERSION = "0.5.31"
# Don't forget to update version in moya/__init__.py

classifiers = [
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3.4"
]

long_desc = """Web development framework"""

setup(name='moya',
      version=VERSION,
      description="web development platform",
      long_description=long_desc,
      zip_safe=False,
      license="MIT",
      author="Will McGugan",
      author_email="will@willmcgugan.com",
      url="http://www.moyaproject.com",

      entry_points={
          "console_scripts": [
              'moya = moya.command.app:main',
              'moya-pm = moya.command.moyapi:main',
              'moya-srv = moya.command.moyasrv:main',
          ]
      },
      scripts=['scripts/moya-workon'],

      platforms=['any'],
      packages=find_packages(),
      include_package_data=True,
      exclude_package_data={'': ['_*', 'docs/*']},

      classifiers=classifiers,
      install_requires=['pyparsing==2.0.5',
                        'webob',
                        'sqlalchemy',
                        'pytz',
                        'pygments',
                        'fs >= 0.5.4',
                        'iso8601',
                        'babel==1.3',
                        'postmarkup',
                        'polib',
                        'pillow==2.9.0',
                        'pymysql',
                        'passlib',
                        'commonmark',
                        'requests',
                        'lxml',
                        'colorama',
                        'premailer',
                        'watchdog',
                        'bleach',
                        'beautifulsoup4'],
      extras_require={
          ':sys_platform=="linux2" or sys_platform=="linux3"': ["pyinotify"],
          'dev': ['notify2']
      },
      setup_requires=["setuptools_git >= 0.3"]
      )



from __future__ import unicode_literals

from moya import loggingconf

from fs.osfs import OSFS

import os.path
from os.path import join


class TestLoggingConf(object):

    def setUp(self):
        self.path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'loggingconf')
        print(self.path)

    def tearDown(self):
        pass

    def test_init(self):
        """Test reading logging conf"""
        loggingconf.init_logging(join(self.path, 'logging.ini'))
        loggingconf.init_logging(join(self.path, 'extend.ini'))

    def test_fs(self):
        """test reading logging from fs"""
        fs = OSFS(self.path)
        loggingconf.init_logging_fs(fs, 'logging.ini')
        loggingconf.init_logging_fs(fs, 'extend.ini')



"""
Various functions to interface templates with moya

"""
from __future__ import unicode_literals
from ..compat import string_types


class Interface(object):

    def __init__(self, archive, context):
        self.archive = archive
        self.context = context
        self.urls = context['.urls']
        self.settings = self.archive.settings

    def render(self, obj, target='html'):
        if isinstance(obj, string_types):
            return obj
        if hasattr(obj, 'moya_render'):
            if hasattr(obj, 'moya_render_targets') and target not in obj.moya_render_targets:
                return ''
            return obj.moya_render(self.archive, self.context, target=target)



"""

File downloader with progress bar and download speed

"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division

from ..console import Console
from ..progress import Progress

import hashlib
import requests
from time import time


class DownloaderError(Exception):
    pass


def _filesize(size):
    try:
        size = int(size)
    except:
        raise ValueError("filesize requires a numeric value, not {!r}".format(size))
    suffixes = ('kB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB')
    base = 1024
    if size == 1:
        return '1 byte'
    elif size < base:
        return '{:,} bytes'.format(size)

    for i, suffix in enumerate(suffixes):
        unit = base ** (i + 2)
        if size < unit:
            return "{:,.01f} {}".format((base * size / unit), suffix)
    return "{:,.01f} {}".format((base * size / unit), suffix)


def download(url,
             store_file,
             filename=None,
             console=None,
             chunk_size=1024 * 16,
             auth=None,
             verify_ssl=True,
             msg="contacting server"):
    """Download a url and render a progress bar"""
    if console is None:
        console = Console()

    if filename is None:
        filename = url.rsplit('/')[-1]

    console.show_cursor(False)
    try:
        progress_width = 24
        progress = Progress(console, '{}'.format(filename), width=progress_width, vanish=False)
        progress.update(0, msg=msg)
        try:
            response = requests.get(url, stream=True, auth=auth, verify=verify_ssl)
            start = time()
            length = response.headers.get('content-length')
            if response.status_code != 200:
                raise DownloaderError('downloader received bad status code ({})'.format(response.status_code))

            m = hashlib.md5()
            bytes_read = 0
            if length is None:
                console('downloading {}'.format(filename))
                console.flush()
                for data in response.iter_content(chunk_size):
                    store_file.write(data)
                    m.update(data)
                console.nl()
            else:
                length = int(length)
                progress.set_num_steps(length)
                for data in response.iter_content(chunk_size):
                    store_file.write(data)
                    m.update(data)
                    bytes_read += len(data)

                    bytes_per_second = int(float(bytes_read) / (time() - start))
                    speed = "{filename}".format(filename=filename, speed=(_filesize(bytes_per_second) + "/s").ljust(16))

                    progress.step(len(data), msg=speed)
        finally:
            progress.done()
    finally:
        console.show_cursor(True)

    return m.hexdigest()



from __future__ import unicode_literals
from ..elements.elementbase import LogicElement, Attribute
from ..filter import MoyaFilter, MoyaFilterExpression


class Filter(LogicElement):
    """
    Define a [i]filter[/i], which may be used in expressions.

    Here's an example of a filter:

    [code xml]
    <filter name="repeat">
        <return-str>${str:value * count}</return-str>
    </filter>
    [/code]

    And here is how you might use it in an expression:

    [code xml]
    <echo>${"beetlejuice "|repeat(count=3)}</echo>
    [/code]
    """

    class Help:
        synopsis = "define a filter"

    name = Attribute("Filter name", required=True)
    value = Attribute("Value name", default="value", required=False)
    expression = Attribute("Expression", type="function", required=False, default=None)
    missing = Attribute("Allow missing values?", type="boolean", default=False, required=False)

    def lib_finalize(self, context):
        validator = None

        for signature in self.children('signature'):
            validator = signature.get_validator(context)

        expression = self.expression(context)
        value_name = self.value(context)
        allow_missing = self.missing(context)
        if expression is not None:
            _filter = MoyaFilterExpression(expression, value_name, allow_missing=allow_missing)
        else:
            _filter = MoyaFilter(self.lib, self.libid, value_name, allow_missing=allow_missing, validator=validator)
        self.lib.register_filter(self.name(context), _filter)
